#include        <stdio.h>
#include        <errno.h>
#include        <X11/Xlib.h>
#include        <X11/XWDFile.h>

/*
 * Get the XColors of all pixels in image - returns # of colors
 */
static int Get_XColors(dpy, win_info, colors)
     Display *dpy;
     XWindowAttributes *win_info;
     XColor **colors;
{
    int i, ncolors;

    if (!win_info->colormap)
        return(0);

    if (win_info->visual->class == TrueColor ||
        win_info->visual->class == DirectColor)
        return(0);    /* XXX punt for now */

    ncolors = win_info->visual->map_entries;
    if (!(*colors = (XColor *) malloc (sizeof(XColor) * ncolors)))
      fprintf (stderr, "xsnap:  out of memory attempting to allocate colors\n");

    for (i=0; i<ncolors; i++)
      (*colors)[i].pixel = i;

    XQueryColors(dpy, win_info->colormap, *colors, ncolors);
    
    return(ncolors);
}

/*
 * Determine the pixmap size.
 */

static int Image_Size(image, format)
     XImage *image;
     int format;
{
    if (format != ZPixmap)
      return(image->bytes_per_line * image->height * image->depth);

    return(image->bytes_per_line * image->height);
}

static int _swapshort (bp, n)
    register char *bp;
    register unsigned n;
{
    register char c;
    register char *ep = bp + n;

    while (bp < ep) {
        c = *bp;
        *bp = *(bp + 1);
        bp++;
        *bp++ = c;
    }
}

static int _swaplong (bp, n)
    register char *bp;
    register unsigned n;
{
    register char c;
    register char *ep = bp + n;
    register char *sp;

    while (bp < ep) {
        sp = bp + 3;
        c = *sp;
        *sp = *bp;
        *bp++ = c;
        sp = bp + 1;
        c = *sp;
        *sp = *bp;
        *bp++ = c;
        bp += 2;
    }
}

int save_as_xwd_file (display, win_info, pixmap, filename)
     Display *display;
     XWindowAttributes *win_info;
     Pixmap pixmap;
     char *filename;
{
    unsigned long swaptest = 1;
    XColor *colors;
    unsigned buffer_size;
    int win_name_size;
    int header_size;
    int ncolors, i;
    char *win_name;
    XImage *image;
    int absx, absy, x, y;
    int dwidth, dheight;
    int bw;
    XWDFileHeader header;
    FILE *output;
    unsigned int width  = win_info->width, 
                 height = win_info->height, 
                 depth  = win_info->depth;

    image = XGetImage (display, pixmap, 0, 0, width, height, AllPlanes, ZPixmap);

    if (!image) {
        fprintf (stderr, "xsnap:  unable to get image at %dx%d+%d+%d\n",
                 width, height, 0, 0);
        exit (1);
    }

    /*
     * Find the output file
     */
    output = fopen (filename, "w+");

    /*
     * Determine the pixmap size.
     */
    buffer_size = Image_Size(image, ZPixmap);

    ncolors = Get_XColors(display, win_info, &colors);

    /*
     * Calculate header size.
     */
    header_size = sizeof(header);

    /*
     * Write out header information.
     */
    header.header_size = (xwdval) header_size;
    header.file_version = (xwdval) XWD_FILE_VERSION;
    header.pixmap_format = (xwdval) ZPixmap;
    header.pixmap_depth = (xwdval) image->depth;
    header.pixmap_width = (xwdval) image->width;
    header.pixmap_height = (xwdval) image->height;
    header.xoffset = (xwdval) image->xoffset;
    header.byte_order = (xwdval) image->byte_order;
    header.bitmap_unit = (xwdval) image->bitmap_unit;
    header.bitmap_bit_order = (xwdval) image->bitmap_bit_order;
    header.bitmap_pad = (xwdval) image->bitmap_pad;
    header.bits_per_pixel = (xwdval) image->bits_per_pixel;
    header.bytes_per_line = (xwdval) image->bytes_per_line;
    header.visual_class = (xwdval) win_info->visual->class;
    header.red_mask = (xwdval) win_info->visual->red_mask;
    header.green_mask = (xwdval) win_info->visual->green_mask;
    header.blue_mask = (xwdval) win_info->visual->blue_mask;
    header.bits_per_rgb = (xwdval) win_info->visual->bits_per_rgb;
    header.colormap_entries = (xwdval) win_info->visual->map_entries;
    header.ncolors = ncolors;
    header.window_width = (xwdval) win_info->width;
    header.window_height = (xwdval) win_info->height;

    header.window_x = 0;
    header.window_y = 0;
    header.window_bdrwidth = (xwdval) win_info->border_width;

    if (*(char *) &swaptest) {
      _swaplong((char *) &header, sizeof(header));
      for (i = 0; i < ncolors; i++) {
        _swaplong((char *) &colors[i].pixel, sizeof(long));
        _swapshort((char *) &colors[i].red, 3 * sizeof(short));
      }
    }

    (void) fwrite((char *)&header, sizeof(header), 1, output);

    /*
     * Write out the color maps, if any
     */
    (void) fwrite((char *) colors, sizeof(XColor), ncolors, output);

    /*
     *    This copying of the bit stream (data) to a file is to be replaced
     *  by an Xlib call which hasn't been written yet.  It is not clear
     *  what other functions of xwd will be taken over by this (as yet)
     *  non-existant X function.
     */
    (void) fwrite(image->data, (int) buffer_size, 1, output);

    /*
     * close the file
     */
    if (output != (stdout))
      fclose (output);

    /*
     * free the color buffer.
     */
    if(ncolors > 0) free(colors);

    /*
     * Free image
     */
    XDestroyImage(image);

    return 0;
}


#ifdef HAS_XPM

#include        <X11/xpm.h>

int save_as_xpm_file (display, win_info, pixmap, filename)
     Display *display;
     XWindowAttributes *win_info;
     Pixmap pixmap;
     char *filename;
{
    XpmAttributes     attributes;

    /* save the image in a file */
    attributes.colormap  = win_info->colormap;
    attributes.width     = win_info->width;
    attributes.height    = win_info->height;
    attributes.depth     = win_info->depth;
    attributes.valuemask = XpmColormap | XpmSize | XpmDepth;

    if(XpmSuccess != XpmWriteFileFromPixmap(display, filename, pixmap, NULL, &attributes)) {
         fprintf(stderr, "error when save pixmap to file.\n");
         return 1;
    }

    return 0;
}
#endif
