/*
 * hxwd.c, version 1.0, Sun May 25 18:10:29 EDT 1997
 *
 * author: tianhua han, tian@aluxpo.micro.lucent.com
 *                      than@lucent.com
 *
 * objective: to snap a huge window bigger than a screen.
 *
 */

#include        <stdio.h>
#include        <errno.h>
#include        <X11/Xlib.h>
#include        <X11/cursorfont.h>

extern int save_as_xwd_file( Display*, XWindowAttributes*, Pixmap pixmap, char*);
extern int save_as_xpm_file( Display*, XWindowAttributes*, Pixmap pixmap, char*);

/* this is for tvtwm and swm, because they make use of virtual root window */ 
extern Window GetVRoot(Display*, int);  

static void usage(char *prog_name)
{
   fprintf(stderr, 
           "usage: %s [-w window_id | -s delay] [-f filename] [-x x_delay] [-g format] [-n] [-v]\n",
            strrchr(prog_name,'/')==NULL? prog_name : (char *)(strrchr(prog_name,'/')+1));
   fprintf(stderr, "   -w window_id  target window id.\n");
   fprintf(stderr, "   -s delay      number of seconds before the window grab. this\n");
   fprintf(stderr, "                 gives you time to locate the target in another\n");
   fprintf(stderr, "                 screen or to pop up another window. default 0.\n");
   fprintf(stderr, "   -x x_delay    number of seconds delayed after the target window\n");
   fprintf(stderr, "                 pops up. this gives time for x server to redraw\n");
   fprintf(stderr, "                 the target window. default 1. for slow x server,\n");
   fprintf(stderr, "                 this should be longer.\n");
   fprintf(stderr, "   -f filename   file where the image is to be saved.\n");
   fprintf(stderr, "                 default \"./junk\"\n");
   fprintf(stderr, "   -g format     \"xpm\" for xpm format, \"xwd\" for xwd format.\n");
   fprintf(stderr, "                 default \"xwd\".\n");
   fprintf(stderr, "   -n            disable compension for window manager frame.\n");
   fprintf(stderr, "                 the compension is enabled by default.\n");
   fprintf(stderr, "   -v            give verbose output.\n");
   exit(1);
}

/*
 * Routine to let user select a window using the mouse
 * (based on xwininfo)
 */

static Window Select_Window(dpy, screen)
    Display *dpy;
    int     screen;
{
    int status;
    Cursor cursor;
    XEvent event;
    Window target_win = None, 
           root = GetVRoot(dpy, screen);

    /* Make the target cursor */
    cursor = XCreateFontCursor(dpy, XC_crosshair);

    /* Grab the pointer using target cursor, letting it room all over */
    status = XGrabPointer(dpy, root, False,
                        ButtonPressMask|ButtonReleaseMask, GrabModeSync,
                        GrabModeAsync, root, cursor, CurrentTime);
    if (status != GrabSuccess) {
       fprintf(stderr, "Can't grab the mouse.\n");
       exit(1);
    }

    /* Let the user select a window... */
    while (target_win == None) {
       /* allow one more event */
       XAllowEvents(dpy, SyncPointer, CurrentTime);
       XWindowEvent(dpy, root, ButtonPressMask|ButtonReleaseMask, &event);
       switch (event.type) {
          case ButtonPress:
             if (target_win == None) {
                 target_win = event.xbutton.subwindow; /* window selected */
                     if (target_win == None) target_win = root;
             }
             break;
          /* discard ButtonRelease event */
       }
    } 

    XUngrabPointer(dpy, CurrentTime);      /* Done with pointer */

    return(target_win);
}


#define XWDFORMAT 0
#define XPMFORMAT 1

void
main(int argc, char **argv)
{
    Display           *dpy    = XOpenDisplay(NULL);
    int               screen  = DefaultScreen(dpy);
    Window            win     = None, junkwin; 

    /* under olvwm, ab_root equal to root while not under tvtwm. */
    Window            ab_root = DefaultRootWindow(dpy),   /* absolute root window */ 
                      root    = GetVRoot(dpy, screen);    /* virtual root window */
/*
    int               depth   = XDefaultDepth(dpy, screen);
    GC                gc      = XDefaultGC(dpy, screen);
*/
    GC                gc;
    Pixmap            pixmap;
    XGCValues         gc_value;
    unsigned long     gc_mask;
    XWindowAttributes win_attr;
    int               i, j, 
                      delay   = 0,
                      x_delay = 1, 
                      format  = XWDFORMAT,
                      xsave, ysave, 
                      rx,    ry,
                      screen_width, screen_height,
                      enable_comp = 1,
                      x_offset    = 0,
                      y_offset    = 0,
                      verbose     = 0;
    char              *filename="./junk";
    XWindowChanges    values;
    unsigned int      value_mask;
 
    if(argc == -1 ) {
        /* this message is for strings(1). */
        printf("Here!!!.\n");
        printf("the source is in tian's pub/src. tian@aluxpo.micro.lucent.com, x7668.\n");
    }

    for(i=0; i<argc; i++) {
        char *arg = argv[i];
        if(arg[0] == '-') {
              switch (arg[1]) {
                 case 'w': if(++i >= argc) usage(argv[0]);
                           win = strtol(argv[i], (char **)NULL, 0);
                           continue;
                 case 's': if(++i >= argc) usage(argv[0]);
                           delay = atoi(argv[i]);
                           continue;
                 case 'x': if(++i >= argc) usage(argv[0]);
                           x_delay = atoi(argv[i]);
                           continue;
                 case 'f': if(++i >= argc) usage(argv[0]);
                           filename = argv[i];
                           continue;
                 case 'g': if(++i >= argc) usage(argv[0]);
                           if ( ! strcmp(argv[i], "xwd"))      
                              format = XWDFORMAT;
                           else if ( ! strcmp(argv[i], "xpm")) 
#ifdef HAS_XPM
                              format = XPMFORMAT;
#else
                           {
                                   
                              fprintf(stderr, 
                                      "%s: xpm isn't built in, ask your admininstrator.\n",
                                      argv[0]);
                              exit(1);
                           } 
#endif
                           else {
                                   fprintf(stderr, "invalid format\n");
                                   usage(argv[0]);
                           }
                           continue;
                 case 'n': enable_comp=0;
                           continue;
                 case 'v': verbose = 1;
                           continue;
                 default : usage(argv[0]);
              }
        }
    }

    /* If no window selected on command line, let user pick one by mouse. */
    if (!win) {
          printf("\n");
          printf("%s: Please select the window about which you would\n", argv[0]);
          printf("        like to grab by click the mouse in that window.\n");
    
          /* in case the user want to move to another screen, give him a litte time. */
          sleep (delay);

          /* get the window interactively */
          win = Select_Window(dpy, screen);

          /* based on xwininfo */
          if (win) {
              Window root;
              int dummyi;
              unsigned int dummy;
              if (XGetGeometry (dpy, win, &root, &dummyi, &dummyi,
                                &dummy, &dummy, &dummy, &dummy) && win != root)
                     win = XmuClientWindow (dpy, win);
          }
    }

    /* because it will take a while, let's grab the input to avoid messing up. */ 
    XBell(dpy, 100);
    if (XGrabKeyboard (dpy, root, False, GrabModeAsync, GrabModeAsync, CurrentTime)
        != GrabSuccess) {
             XBell(dpy, 100);
             fprintf(stderr, "warning: cannot grab the keyboard. \n");
    }
    if (XGrabPointer (dpy, root,
                      False, 0, GrabModeAsync, GrabModeAsync, None, 
                      XCreateFontCursor(dpy, XC_watch), 
                      CurrentTime)        
       != GrabSuccess) {
             XBell(dpy, 100);
             fprintf(stderr, "warning: cannot grab the pointer. \n");
    }

    if(verbose) {
          printf("ab_root_id=0x%lx\n", ab_root);
          printf("root_id=0x%lx\n", root);
          printf("window_id=0x%lx\n", win);
          printf("delay=%d\n", delay);
          printf("x_delay=%d\n", x_delay);
          printf("filename=%s\n", filename);
          if( format == XWDFORMAT) printf("use xwd format\n");
          if( format == XPMFORMAT) printf("use xpm format\n");
          if( enable_comp ) printf("window manager frame compensation enabled.\n");
          else              printf("window manager frame compensation disabled.\n");
    }

    if (XGetWindowAttributes(dpy, win, &win_attr) == 0) {
            fprintf(stderr, "error when inqury the window postion.\n");
            exit(1);
    }

    if (enable_comp) {
            x_offset = win_attr.x;
            y_offset = win_attr.y;
    } 

    /* reserver a pixmap */
    pixmap = XCreatePixmap(dpy, root, win_attr.width, win_attr.height, win_attr.depth);

    /* set the gc */
    /* gc_mask  = GCSubwindowMode | GCClipMask; */
    gc_mask  = GCSubwindowMode;
    gc_value.subwindow_mode = IncludeInferiors;
    gc = XCreateGC(dpy, win, gc_mask, &gc_value);

    /* save the window's old position */
    /*
     * get window position, compensating for decorations
     * (based on xwininfo.c)
     */
    (void) XTranslateCoordinates(dpy, win, root, /*or win_attr.root which is ab_root*/
               -win_attr.border_width, -win_attr.border_width,
             &rx, &ry, &junkwin);
    /* win_attr.x and win_attr.y are the window manager's wrapper window's frame */
    xsave = rx - x_offset;    
    ysave = ry - y_offset;
    
    if(verbose) {
         printf("rx=%d, ry=%d\n", rx, ry);
         printf("win_attr.root=0x%lx\n", win_attr.root);
         printf("win_attr.x=%d, win_attr.y=%d\n", win_attr.x, win_attr.y);
         printf("win_attr.border_width=%d\n", win_attr.border_width);
         printf("x_offset=%d, y_offset=%d\n", x_offset, y_offset);
         printf("xsave=%d, ysave=%d\n", xsave, ysave);
    }

    /* raise up the window, but this will take effect only after Xsync */
    XMapRaised(dpy, win);
    
    /* 
     * check whether the original window is in a state to be watched? 
     * note: olvwm and tvtwm treat virtual window -- the window not in present
     *       screen -- differently. under olvwm it is viewable while in tvtwm
     *       it is not.
     */
/*
    if (win_attr.map_state != IsViewable) {
        fprintf(stderr,
                "The window you wish to look at is not in a viewable state.\n");
        fprintf(stderr,
                "Perhaps it is iconified or not mapped.\n");
        exit(1);
    }
*/
    
    /* get the dimesion of the screen */
    screen_width=DisplayWidth (dpy, screen);
    screen_height=DisplayHeight (dpy, screen);

    if(verbose) 
        printf("screen_width=%d,screen_height=%d.\n", screen_width,screen_height);

/**********************************************************

http://www.cis.ohio-state.edu/hypertext/faq/usenet/x-faq/part7/faq-doc-20.html

168) Why do I get a BadMatch error when calling XGetImage?

The BadMatch error can occur if the specified rectangle goes off the edge of 
the screen. If you don't want to catch the error and deal with it, you can take
the following steps to avoid the error:

1) Make a pixmap the same size as the rectangle you want to capture.
2) Clear the pixmap to background using XFillRectangle.
3) Use XCopyArea to copy the window to the pixmap.
        [Whoa! this answer is currently under reexamination.]
4) If you get a NoExpose event, the copy was clean. Use XGetImage to grab the
image from the pixmap.
5) If you get one or more GraphicsExpose events, the copy wasn't clean, and 
the x/y/width/height members of the GraphicsExpose event structures tell you 
the parts of the pixmap which aren't good.
6) Get rid of the pixmap; it probably takes a lot of memory.

[10/92; thanks to Oliver Jones (oj@pictel.com)]

Comment:

the above process helped me get rid of the error message, but only grab the
exposed part of image. i have tried on several kinds of x servers, including
NCD xterminal, eXceed running on pc (pentium with win95). none of their Backing
Store as X11R6.1 documents said, that is, they won't give my XCopyArea the
content of the obscued part of the window. my conclusion is: we have to expose
the part of the target window before we snap it. my approach is totally 3 steps
to scan a huge window. for a small window (if i can be covered be one screen),
only one step is needed.
                                           -tian

        +----------------+---------------+----+
        | step 1                              |     
        |                                     |
        |  screen 0,0      screen 1,0         |
        |                                     |
        |                                     |
        +                +               +    +
        |                                     |     
        |                                     |
        |  screen 0,1      screen 1,1         |
        |                                     |
        |                                     |
        +----------------+---------------+----+
        | step 2                         | step 3   
        |                                |    |
        |                                |    |
        +--------------------------------|----+

************************************************************************/


    /* step 1 */
    for (j=0; j<(win_attr.height-1)/screen_height; j++) {
        for (i=0; i<(win_attr.width-1)/screen_width; i++) {
 
             if(verbose) printf("operating in screen (%d,%d).\n", i, j);

             /* shift the window to let the part to be copyied exposed */
             values.x = -i*screen_width  - x_offset;
             values.y = -j*screen_height - y_offset;
             value_mask = CWX | CWY;
             if (XReconfigureWMWindow(dpy, win, screen, value_mask, &values) == 0) {
                 fprintf(stderr, 
                         "%s: error when shift the window to screen, (-%d,-%d).\n",
                         argv[0], i, j);
                 exit(1);
             }

             /* must xsync it; otherwise the XMapRaised will be delayed */
             XSync(dpy, False);

             /* wait a few seconds to let the x server refresh the window */
             sleep(x_delay);

             /* copy one screen */
             XCopyArea(dpy, win, pixmap, gc, i*screen_width, j*screen_height, 
                   screen_width, screen_height, i*screen_width, j*screen_height);

         } 

         if(verbose) printf("operating in screen (%d,%d).\n", i, j);

         /* copy the last part of the window in this row 
          * with only partial width of the screen.
          *
          * shift the window to let the part to be copyied exposed 
          */
         values.x = -i*screen_width  - x_offset;
         values.y = -j*screen_height - y_offset;
         value_mask = CWX | CWY;
         if (XReconfigureWMWindow(dpy, win, screen, value_mask, &values) == 0) {
                 fprintf(stderr, 
                         "%s: error when shift the window to screen, (-%d,-%d).\n",
                         argv[0], i, j);
                 exit(1);
        }

        /* must xsync it; otherwise the XMapRaised will be delayed */
        XSync(dpy, False);

        /* wait a few seconds to let the x server refresh the window */
        sleep(x_delay);
 
        /* copy the exposed area */
        XCopyArea(dpy, win, pixmap, gc, i*screen_width, j*screen_height, 
                  win_attr.width - i*screen_width, /* partial width */
                  screen_height,                   /* whole height */
                  i*screen_width, j*screen_height);

    }

    /* step 2, copy the last row with partial height of the screen */
    for (i=0; i<(win_attr.width-1)/screen_width; i++) {
        
        if(verbose) printf("operating in screen (%d,%d).\n", i, j);

        /* shift the window to let the part to be copyied exposed */
        values.x = -i*screen_width  - x_offset;
        values.y = -j*screen_height - y_offset;
        value_mask = CWX | CWY;
        if (XReconfigureWMWindow(dpy, win, screen, value_mask, &values) == 0) {
                 fprintf(stderr, 
                         "%s: error when shift the window to screen, (-%d,-%d).\n",
                         argv[0], i, j);
                 exit(1);
        }

        /* must xsync it; otherwise the XMapRaised will be delayed */
        XSync(dpy, False);

        /* wait a few seconds to let the x server refresh the window */
        sleep(x_delay);

        /* copy one partial screen */
        XCopyArea(dpy, win, pixmap, gc, i*screen_width, j*screen_height, 
                   screen_width,                       /* whole width */
                   win_attr.height - j*screen_height,  /* partial height */
                   i*screen_width, j*screen_height);

    } 
 
    /* step 3, process the last corner, right-bottom */
    if(verbose) printf("operating in screen (%d,%d).\n", i, j);

    /* shift the window to let the part to be copyied exposed */
    values.x = -i*screen_width  - x_offset;
    values.y = -j*screen_height - y_offset;
    value_mask = CWX | CWY;
    if (XReconfigureWMWindow(dpy, win, screen, value_mask, &values) == 0) {
                 fprintf(stderr, 
                         "%s: error when shift the window to screen, (-%d,-%d).\n",
                         argv[0], i, j);
                 exit(1);
    }

    /* must xsync it; otherwise the XMapRaised will be delayed */
    XSync(dpy, False);

    /* wait a few seconds to let the x server refresh the window */
    sleep(x_delay);
 
    /* copy the last corner, right-bottom */
    XCopyArea(dpy, win, pixmap, gc, i*screen_width, j*screen_height, 
              win_attr.width  - i*screen_width,  /* partial width */
              win_attr.height - j*screen_height, /* partial height */
              i*screen_width, j*screen_height);
    

    /* restore the window positiion. */
    values.x = xsave;
    values.y = ysave;
    value_mask = CWX | CWY;
    if (XReconfigureWMWindow(dpy, win, screen, value_mask, &values) == 0) {
           fprintf(stderr, "%s: error when move window back.\n", argv[0]);
           exit(1);
    }
    
    /* must sync, otherwise XReconfigureWMWindow won't take effect */
    XSync(dpy, False);
    XBell(dpy, 100);

    /* now restore the user input */
    XUngrabKeyboard(dpy, CurrentTime);
    XUngrabPointer(dpy,  CurrentTime);

    if(verbose) printf("saving file to \"%s\". please wait.\n", filename);

    if (format == XWDFORMAT)
          save_as_xwd_file (dpy, &win_attr, pixmap, filename);

#ifdef HAS_XPM
    if (format == XPMFORMAT) 
          save_as_xpm_file (dpy, &win_attr, pixmap, filename);
#endif

    /* free the pixmap */
    XFreePixmap(dpy, pixmap);
}
